@@ -1,4 +1,4 @@
-#define SEQUEL_PG_VERSION_INTEGER 11400
+#define SEQUEL_PG_VERSION_INTEGER 11802
 
 #include <string.h>
 #include <stdio.h>
@@ -24,6 +24,10 @@
 #define RARRAY_AREF(a, i) (RARRAY_PTR(a)[i])
 #endif
 
+#ifndef HAVE_RB_HASH_NEW_CAPA
+#define rb_hash_new_capa(_) rb_hash_new()
+#endif
+
 #define ntohll(c) ((uint64_t)( \
     (((uint64_t)(*((unsigned char*)(c)+0)))<<56LL) | \
     (((uint64_t)(*((unsigned char*)(c)+1)))<<48LL) | \
@@ -65,14 +69,26 @@
 #define SPG_YIELD_MKV_HASH_GROUPS 11
 #define SPG_YIELD_KMV_HASH_GROUPS 12
 #define SPG_YIELD_MKMV_HASH_GROUPS 13
+#define SPG_YIELD_COLUMN_ARRAY 14
+#define SPG_YIELD_COLUMNS_ARRAY 15
+#define SPG_YIELD_FIRST_ARRAY 16
+#define SPG_YIELD_ARRAY_ARRAY 17
+#define SPG_YIELD_COLUMN_SET 18
+#define SPG_YIELD_COLUMNS_SET 19
+#define SPG_YIELD_FIRST_SET 20
+#define SPG_YIELD_ARRAY_SET 21
+#define SPG_YIELD_ALL 22
+#define SPG_YIELD_ALL_MODEL 23
 
 /* External functions defined by ruby-pg */
 PGconn* pg_get_pgconn(VALUE);
 PGresult* pgresult_get(VALUE);
 int pg_get_result_enc_idx(VALUE);
+VALUE pgresult_stream_any(VALUE self, int (*yielder)(VALUE, int, int, void*), void* data);
 
 static int spg_use_ipaddr_alloc;
 static int spg_use_pg_get_result_enc_idx;
+static int spg_use_pg_stream_any;
 
 static VALUE spg_Sequel;
 static VALUE spg_PGArray;
@@ -89,11 +105,19 @@ static VALUE spg_vmasks6;
 static VALUE spg_sym_utc;
 static VALUE spg_sym_local;
 static VALUE spg_sym_map;
+static VALUE spg_sym_map_array;
+static VALUE spg_sym_map_set;
 static VALUE spg_sym_first;
+static VALUE spg_sym_first_array;
+static VALUE spg_sym_first_set;
 static VALUE spg_sym_array;
+static VALUE spg_sym_array_array;
+static VALUE spg_sym_array_set;
 static VALUE spg_sym_hash;
 static VALUE spg_sym_hash_groups;
 static VALUE spg_sym_model;
+static VALUE spg_sym_all;
+static VALUE spg_sym_all_model;
 static VALUE spg_sym__sequel_pg_type;
 static VALUE spg_sym__sequel_pg_value;
 
@@ -169,7 +193,7 @@ static ID spg_id_family;
 static ID spg_id_addr;
 static ID spg_id_mask_addr;
 
-#if HAVE_PQSETSINGLEROWMODE
+#ifdef HAVE_PQSETSINGLEROWMODE
 static ID spg_id_get_result;
 static ID spg_id_clear;
 static ID spg_id_check;
@@ -197,10 +221,10 @@ static int enc_get_index(VALUE val) {
   } while(0)
 
 static VALUE
-pg_text_dec_integer(char *val, int len)
+pg_text_dec_integer(char *val, size_t len)
 {
   long i;
-  int max_len;
+  size_t max_len;
 
   if( sizeof(i) >= 8 && FIXNUM_MAX >= 1000000000000000000LL ){
     /* 64 bit system can safely handle all numbers up to 18 digits as Fixnum */
@@ -255,7 +279,7 @@ pg_text_dec_integer(char *val, int len)
 
 static VALUE spg__array_col_value(char *v, size_t length, VALUE converter, int enc_index, int oid, VALUE db);
 
-static VALUE read_array(int *index, char *c_pg_array_string, int array_string_length, VALUE buf, VALUE converter, int enc_index, int oid, VALUE db) {
+static VALUE read_array(int *index, char *c_pg_array_string, long array_string_length, VALUE buf, VALUE converter, int enc_index, int oid, VALUE db) {
   int word_index = 0;
   char *word = RSTRING_PTR(buf);
 
@@ -351,7 +375,7 @@ static VALUE read_array(int *index, char *c_pg_array_string, int array_string_le
   return array;
 }
 
-static VALUE check_pg_array(int* index, char *c_pg_array_string, int array_string_length) {
+static VALUE check_pg_array(int* index, char *c_pg_array_string, long array_string_length) {
   if (array_string_length == 0) {
     rb_raise(rb_eArgError, "unexpected PostgreSQL array format, empty");
   } else if (array_string_length == 2 && c_pg_array_string[0] == '{' && c_pg_array_string[0] == '}') {
@@ -382,7 +406,7 @@ static VALUE parse_pg_array(VALUE self, VALUE pg_array_string, VALUE converter)
   /* convert to c-string, create additional ruby string buffer of
    * the same length, as that will be the worst case. */
   char *c_pg_array_string = StringValueCStr(pg_array_string);
-  int array_string_length = RSTRING_LEN(pg_array_string);
+  long array_string_length = RSTRING_LEN(pg_array_string);
   int index = 1;
   VALUE ary;
 
@@ -532,7 +556,7 @@ static VALUE spg_timestamp(const char *s, VALUE self, size_t length, int tz) {
   }
 
   if (remaining < 19) {
-    return spg_timestamp_error(s, self, "unexpected timetamp format, too short");
+    return spg_timestamp_error(s, self, "unexpected timestamp format, too short");
   }
 
   year = parse_year(&p, &remaining);
@@ -877,7 +901,7 @@ static VALUE spg__array_col_value(char *v, size_t length, VALUE converter, int e
       break;
     case 17: /* bytea */
       bi.blob_string = (char *)PQunescapeBytea((unsigned char*)v, &bi.length);
-      rv = rb_ensure(spg_create_Blob, (VALUE)&bi, (VALUE(*)())PQfreemem, (VALUE)bi.blob_string);
+      rv = rb_ensure(spg_create_Blob, (VALUE)&bi, (VALUE(*)(VALUE))PQfreemem, (VALUE)bi.blob_string);
       break;
     case 20: /* integer */
     case 21:
@@ -1010,12 +1034,12 @@ static int spg_timestamp_info_bitmask(VALUE self) {
   return tz;
 }
 
-static VALUE spg__col_value(VALUE self, PGresult *res, long i, long j, VALUE* colconvert, int enc_index) {
+static VALUE spg__col_value(VALUE self, PGresult *res, int i, int j, VALUE* colconvert, int enc_index) {
   char *v;
   VALUE rv;
   int ftype = PQftype(res, j);
   VALUE array_type;
-  VALUE scalar_oid;
+  int scalar_oid;
   struct spg_blob_initialization bi;
 
   if(PQgetisnull(res, i, j)) {
@@ -1029,7 +1053,7 @@ static VALUE spg__col_value(VALUE self, PGresult *res, long i, long j, VALUE* co
         break;
       case 17: /* bytea */
         bi.blob_string = (char *)PQunescapeBytea((unsigned char*)v, &bi.length);
-        rv = rb_ensure(spg_create_Blob, (VALUE)&bi, (VALUE(*)())PQfreemem, (VALUE)bi.blob_string);
+        rv = rb_ensure(spg_create_Blob, (VALUE)&bi, (VALUE(*)(VALUE))PQfreemem, (VALUE)bi.blob_string);
         break;
       case 20: /* integer */
       case 21:
@@ -1249,20 +1273,20 @@ static VALUE spg__col_value(VALUE self, PGresult *res, long i, long j, VALUE* co
   return rv;
 }
 
-static VALUE spg__col_values(VALUE self, VALUE v, VALUE *colsyms, long nfields, PGresult *res, long i, VALUE *colconvert, int enc_index) {
+static VALUE spg__col_values(VALUE self, VALUE v, VALUE *colsyms, long nfields, PGresult *res, int i, VALUE *colconvert, int enc_index) {
   long j;
   VALUE cur;
   long len = RARRAY_LEN(v);
   VALUE a = rb_ary_new2(len);
   for (j=0; j<len; j++) {
     cur = rb_ary_entry(v, j);
-    rb_ary_store(a, j, cur == Qnil ? Qnil : spg__col_value(self, res, i, NUM2LONG(cur), colconvert, enc_index));
+    rb_ary_store(a, j, cur == Qnil ? Qnil : spg__col_value(self, res, i, NUM2INT(cur), colconvert, enc_index));
   }
   return a;
 }
 
-static long spg__field_id(VALUE v, VALUE *colsyms, long nfields) {
-  long j;
+static int spg__field_id(VALUE v, VALUE *colsyms, long nfields) {
+  int j;
   for (j=0; j<nfields; j++) {
     if (colsyms[j] == v) {
       return j;
@@ -1273,7 +1297,7 @@ static long spg__field_id(VALUE v, VALUE *colsyms, long nfields) {
 
 static VALUE spg__field_ids(VALUE v, VALUE *colsyms, long nfields) {
   long i;
-  long j;
+  int j;
   VALUE cur;
   long len = RARRAY_LEN(v);
   VALUE pg_columns = rb_ary_new2(len);
@@ -1286,9 +1310,9 @@ static VALUE spg__field_ids(VALUE v, VALUE *colsyms, long nfields) {
 }
 
 static void spg_set_column_info(VALUE self, PGresult *res, VALUE *colsyms, VALUE *colconvert, int enc_index) {
-  long i;
-  long j;
-  long nfields;
+  int i;
+  int j;
+  int nfields;
   int timestamp_info = 0;
   int time_info = 0;
   VALUE conv_procs = 0;
@@ -1378,10 +1402,10 @@ static void spg_set_column_info(VALUE self, PGresult *res, VALUE *colsyms, VALUE
 }
 
 static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_index, VALUE* colsyms, VALUE* colconvert) {
-  long ntuples;
-  long nfields;
-  long i;
-  long j;
+  int ntuples;
+  int nfields;
+  int i;
+  int j;
   VALUE h;
   VALUE opts;
   VALUE pg_type;
@@ -1403,10 +1427,34 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
         } else if (rb_type(pg_value) == T_ARRAY) {
           type = SPG_YIELD_COLUMNS;
         }
+      } else if (pg_type == spg_sym_map_array) {
+        if (SYMBOL_P(pg_value)) {
+          type = SPG_YIELD_COLUMN_ARRAY;
+        } else if (rb_type(pg_value) == T_ARRAY) {
+          type = SPG_YIELD_COLUMNS_ARRAY;
+        }
+#ifdef HAVE_RB_SET_NEW_CAPA
+      } else if (pg_type == spg_sym_map_set) {
+        if (SYMBOL_P(pg_value)) {
+          type = SPG_YIELD_COLUMN_SET;
+        } else if (rb_type(pg_value) == T_ARRAY) {
+          type = SPG_YIELD_COLUMNS_SET;
+        }
+#endif
       } else if (pg_type == spg_sym_first) {
         type = SPG_YIELD_FIRST;
       } else if (pg_type == spg_sym_array) {
         type = SPG_YIELD_ARRAY;
+      } else if (pg_type == spg_sym_first_array) {
+        type = SPG_YIELD_FIRST_ARRAY;
+      } else if (pg_type == spg_sym_array_array) {
+        type = SPG_YIELD_ARRAY_ARRAY;
+#ifdef HAVE_RB_SET_NEW_CAPA
+      } else if (pg_type == spg_sym_first_set) {
+        type = SPG_YIELD_FIRST_SET;
+      } else if (pg_type == spg_sym_array_set) {
+        type = SPG_YIELD_ARRAY_SET;
+#endif
       } else if ((pg_type == spg_sym_hash || pg_type == spg_sym_hash_groups) && rb_type(pg_value) == T_ARRAY) {
         VALUE pg_value_key, pg_value_value;
         pg_value_key = rb_ary_entry(pg_value, 0);
@@ -1426,6 +1474,10 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
         }
       } else if (pg_type == spg_sym_model && rb_type(pg_value) == T_CLASS) {
         type = SPG_YIELD_MODEL;
+      } else if (pg_type == spg_sym_all_model && rb_type(pg_value) == T_CLASS) {
+        type = SPG_YIELD_ALL_MODEL;
+      } else if (pg_type == spg_sym_all) {
+        type = SPG_YIELD_ALL;
       }
     }
   }
@@ -1434,7 +1486,7 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
     case SPG_YIELD_NORMAL:
       /* Normal, hash for entire row */
       for(i=0; i<ntuples; i++) {
-        h = rb_hash_new();
+        h = rb_hash_new_capa(nfields);
         for(j=0; j<nfields; j++) {
           rb_hash_aset(h, colsyms[j], spg__col_value(self, res, i, j, colconvert, enc_index));
         }
@@ -1461,6 +1513,63 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
         rb_yield(spg__col_values(self, h, colsyms, nfields, res, i, colconvert, enc_index));
       } 
       break;
+    case SPG_YIELD_COLUMN_ARRAY:
+      /* Array containing single column */
+      {
+        VALUE ary = rb_ary_new2(ntuples);
+        j = spg__field_id(pg_value, colsyms, nfields);
+        if (j == -1) {
+          for(i=0; i<ntuples; i++) {
+            rb_ary_store(ary, i, Qnil);
+          } 
+        }
+        else {
+          for(i=0; i<ntuples; i++) {
+            rb_ary_store(ary, i, spg__col_value(self, res, i, j, colconvert, enc_index));
+          } 
+        }
+        rb_yield(ary);
+      }
+      break;
+    case SPG_YIELD_COLUMNS_ARRAY:
+      /* Array containing arrays of columns */
+      {
+        VALUE ary = rb_ary_new2(ntuples);
+        h = spg__field_ids(pg_value, colsyms, nfields);
+        for(i=0; i<ntuples; i++) {
+          rb_ary_store(ary, i, spg__col_values(self, h, colsyms, nfields, res, i, colconvert, enc_index));
+        } 
+        rb_yield(ary);
+      }
+      break;
+#ifdef HAVE_RB_SET_NEW_CAPA
+    case SPG_YIELD_COLUMN_SET:
+      /* Set containing single column */
+      {
+        VALUE set = rb_set_new_capa(ntuples);
+        j = spg__field_id(pg_value, colsyms, nfields);
+        if (j == -1) {
+          rb_set_add(set, Qnil);
+        } else {
+          for(i=0; i<ntuples; i++) {
+            rb_set_add(set, spg__col_value(self, res, i, j, colconvert, enc_index));
+          } 
+        }
+        rb_yield(set);
+      }
+      break;
+    case SPG_YIELD_COLUMNS_SET:
+      /* Set containing arrays of columns */
+      {
+        VALUE set = rb_set_new_capa(ntuples);
+        h = spg__field_ids(pg_value, colsyms, nfields);
+        for(i=0; i<ntuples; i++) {
+          rb_set_add(set, spg__col_values(self, h, colsyms, nfields, res, i, colconvert, enc_index));
+        } 
+        rb_yield(set);
+      }
+      break;
+#endif
     case SPG_YIELD_FIRST:
       /* First column */
       for(i=0; i<ntuples; i++) {
@@ -1477,23 +1586,73 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
         rb_yield(h);
       } 
       break;
+    case SPG_YIELD_FIRST_ARRAY:
+      /* Array of first column */
+      h = rb_ary_new2(ntuples);
+      for(i=0; i<ntuples; i++) {
+        rb_ary_store(h, i, spg__col_value(self, res, i, 0, colconvert, enc_index));
+      } 
+      rb_yield(h);
+      break;
+    case SPG_YIELD_ARRAY_ARRAY:
+      /* Array of arrays of all columns */
+      {
+        VALUE ary = rb_ary_new2(ntuples);
+        for(i=0; i<ntuples; i++) {
+          h = rb_ary_new2(nfields);
+          for(j=0; j<nfields; j++) {
+            rb_ary_store(h, j, spg__col_value(self, res, i, j, colconvert, enc_index));
+          }
+          rb_ary_store(ary, i, h);
+        } 
+        rb_yield(ary);
+      }
+      break;
+#ifdef HAVE_RB_SET_NEW_CAPA
+    case SPG_YIELD_FIRST_SET:
+      /* Array of first column */
+      h = rb_set_new_capa(ntuples);
+      for(i=0; i<ntuples; i++) {
+        rb_set_add(h, spg__col_value(self, res, i, 0, colconvert, enc_index));
+      } 
+      rb_yield(h);
+      break;
+    case SPG_YIELD_ARRAY_SET:
+      /* Array of arrays of all columns */
+      {
+        VALUE set = rb_set_new_capa(ntuples);
+        for(i=0; i<ntuples; i++) {
+          h = rb_ary_new2(nfields);
+          for(j=0; j<nfields; j++) {
+            rb_ary_store(h, j, spg__col_value(self, res, i, j, colconvert, enc_index));
+          }
+          rb_set_add(set, h);
+        } 
+        rb_yield(set);
+      }
+      break;
+#endif
     case SPG_YIELD_KV_HASH:
     case SPG_YIELD_KV_HASH_GROUPS:
       /* Hash with single key and single value */
       {
-        VALUE k, v;
-        h = rb_hash_new();
+        int k, v;
+        VALUE kv, vv;
         k = spg__field_id(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_id(rb_ary_entry(pg_value, 1), colsyms, nfields);
         if(type == SPG_YIELD_KV_HASH) {
+          h = rb_hash_new_capa(ntuples);
           for(i=0; i<ntuples; i++) {
-            rb_hash_aset(h, spg__col_value(self, res, i, k, colconvert, enc_index), spg__col_value(self, res, i, v, colconvert, enc_index));
+            kv = k == -1 ? Qnil : spg__col_value(self, res, i, k, colconvert, enc_index);
+            vv = v == -1 ? Qnil : spg__col_value(self, res, i, v, colconvert, enc_index);
+            rb_hash_aset(h, kv, vv);
           } 
         } else {
-          VALUE kv, vv, a;
+          VALUE a;
+          h = rb_hash_new();
           for(i=0; i<ntuples; i++) {
-            kv = spg__col_value(self, res, i, k, colconvert, enc_index);
-            vv = spg__col_value(self, res, i, v, colconvert, enc_index);
+            kv = k == -1 ? Qnil : spg__col_value(self, res, i, k, colconvert, enc_index);
+            vv = v == -1 ? Qnil : spg__col_value(self, res, i, v, colconvert, enc_index);
             a = rb_hash_lookup(h, kv);
             if(!RTEST(a)) {
               rb_hash_aset(h, kv, rb_ary_new3(1, vv));
@@ -1509,19 +1668,22 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
     case SPG_YIELD_MKV_HASH_GROUPS:
       /* Hash with array of keys and single value */
       {
-        VALUE k, v;
-        h = rb_hash_new();
+        VALUE k, vv;
+        int v;
         k = spg__field_ids(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_id(rb_ary_entry(pg_value, 1), colsyms, nfields);
         if(type == SPG_YIELD_MKV_HASH) {
+          h = rb_hash_new_capa(ntuples);
           for(i=0; i<ntuples; i++) {
-            rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert, enc_index), spg__col_value(self, res, i, v, colconvert, enc_index));
+            vv = v == -1 ? Qnil : spg__col_value(self, res, i, v, colconvert, enc_index);
+            rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert, enc_index), vv);
           } 
         } else {
-          VALUE kv, vv, a;
+          VALUE kv, a;
+          h = rb_hash_new();
           for(i=0; i<ntuples; i++) {
             kv = spg__col_values(self, k, colsyms, nfields, res, i, colconvert, enc_index);
-            vv = spg__col_value(self, res, i, v, colconvert, enc_index);
+            vv = v == -1 ? Qnil : spg__col_value(self, res, i, v, colconvert, enc_index);
             a = rb_hash_lookup(h, kv);
             if(!RTEST(a)) {
               rb_hash_aset(h, kv, rb_ary_new3(1, vv));
@@ -1537,18 +1699,21 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
     case SPG_YIELD_KMV_HASH_GROUPS:
       /* Hash with single keys and array of values */
       {
-        VALUE k, v;
-        h = rb_hash_new();
+        VALUE v, kv;
+        int k;
         k = spg__field_id(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_ids(rb_ary_entry(pg_value, 1), colsyms, nfields);
         if(type == SPG_YIELD_KMV_HASH) {
+          h = rb_hash_new_capa(ntuples);
           for(i=0; i<ntuples; i++) {
-            rb_hash_aset(h, spg__col_value(self, res, i, k, colconvert, enc_index), spg__col_values(self, v, colsyms, nfields, res, i, colconvert, enc_index));
+            kv = k == -1 ? Qnil : spg__col_value(self, res, i, k, colconvert, enc_index);
+            rb_hash_aset(h, kv, spg__col_values(self, v, colsyms, nfields, res, i, colconvert, enc_index));
           } 
         } else {
-          VALUE kv, vv, a;
+          VALUE vv, a;
+          h = rb_hash_new();
           for(i=0; i<ntuples; i++) {
-            kv = spg__col_value(self, res, i, k, colconvert, enc_index);
+            kv = k == -1 ? Qnil : spg__col_value(self, res, i, k, colconvert, enc_index);
             vv = spg__col_values(self, v, colsyms, nfields, res, i, colconvert, enc_index);
             a = rb_hash_lookup(h, kv);
             if(!RTEST(a)) {
@@ -1566,15 +1731,16 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
       /* Hash with array of keys and array of values */
       {
         VALUE k, v;
-        h = rb_hash_new();
         k = spg__field_ids(rb_ary_entry(pg_value, 0), colsyms, nfields);
         v = spg__field_ids(rb_ary_entry(pg_value, 1), colsyms, nfields);
         if(type == SPG_YIELD_MKMV_HASH) {
+          h = rb_hash_new_capa(ntuples);
           for(i=0; i<ntuples; i++) {
             rb_hash_aset(h, spg__col_values(self, k, colsyms, nfields, res, i, colconvert, enc_index), spg__col_values(self, v, colsyms, nfields, res, i, colconvert, enc_index));
           } 
         } else {
           VALUE kv, vv, a;
+          h = rb_hash_new();
           for(i=0; i<ntuples; i++) {
             kv = spg__col_values(self, k, colsyms, nfields, res, i, colconvert, enc_index);
             vv = spg__col_values(self, v, colsyms, nfields, res, i, colconvert, enc_index);
@@ -1592,7 +1758,7 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
     case SPG_YIELD_MODEL:
       /* Model object for entire row */
       for(i=0; i<ntuples; i++) {
-        h = rb_hash_new();
+        h = rb_hash_new_capa(nfields);
         for(j=0; j<nfields; j++) {
           rb_hash_aset(h, colsyms[j], spg__col_value(self, res, i, j, colconvert, enc_index));
         }
@@ -1602,6 +1768,35 @@ static VALUE spg_yield_hash_rows_internal(VALUE self, PGresult *res, int enc_ind
         rb_yield(pg_type);
       }
       break;
+    case SPG_YIELD_ALL_MODEL:
+      {
+        VALUE ary = rb_ary_new2(ntuples);
+        VALUE obj;
+        for(i=0; i<ntuples; i++) {
+          h = rb_hash_new_capa(nfields);
+          for(j=0; j<nfields; j++) {
+            rb_hash_aset(h, colsyms[j], spg__col_value(self, res, i, j, colconvert, enc_index));
+          }
+          obj = rb_obj_alloc(pg_value);
+          rb_ivar_set(obj, spg_id_values, h);
+          rb_ary_store(ary, i, obj);
+        }
+        rb_yield(ary);
+      }
+      break;
+    case SPG_YIELD_ALL:
+      {
+        VALUE ary = rb_ary_new2(ntuples);
+        for(i=0; i<ntuples; i++) {
+          h = rb_hash_new_capa(nfields);
+          for(j=0; j<nfields; j++) {
+            rb_hash_aset(h, colsyms[j], spg__col_value(self, res, i, j, colconvert, enc_index));
+          }
+          rb_ary_store(ary, i, h);
+        }
+        rb_yield(ary);
+      }
+      break;
   }
   return self;
 }
@@ -1619,7 +1814,7 @@ def_spg_yield_hash_rows(1664)
 
 static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
   PGresult *res;
-  long nfields;
+  int nfields;
   int enc_index;
 
   if (!RTEST(rres)) {
@@ -1634,7 +1829,7 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
   else if (nfields <= 64) return spg_yield_hash_rows_64(self, res, enc_index);
   else if (nfields <= 256) return spg_yield_hash_rows_256(self, res, enc_index);
   else if (nfields <= 1664) return spg_yield_hash_rows_1664(self, res, enc_index);
-  else rb_raise(rb_eRangeError, "more than 1664 columns in query (%ld columns detected)", nfields);
+  else rb_raise(rb_eRangeError, "more than 1664 columns in query (%d columns detected)", nfields);
 
   /* UNREACHABLE */
   return self;
@@ -1642,14 +1837,14 @@ static VALUE spg_yield_hash_rows(VALUE self, VALUE rres, VALUE ignore) {
 
 static VALUE spg_supports_streaming_p(VALUE self) {
   return
-#if HAVE_PQSETSINGLEROWMODE
+#ifdef HAVE_PQSETSINGLEROWMODE
   Qtrue;
 #else
   Qfalse;
 #endif
 }
 
-#if HAVE_PQSETSINGLEROWMODE
+#ifdef HAVE_PQSETSINGLEROWMODE
 static VALUE spg_set_single_row_mode(VALUE self) {
   PGconn *conn;
   conn = pg_get_pgconn(self);
@@ -1659,14 +1854,48 @@ static VALUE spg_set_single_row_mode(VALUE self) {
   return Qnil;
 }
 
+struct spg__yield_each_row_stream_data {
+  VALUE self;
+  VALUE *colsyms;
+  VALUE *colconvert;
+  VALUE pg_value;
+  int enc_index;
+  char type;
+};
+
+static int spg__yield_each_row_stream(VALUE rres, int ntuples, int nfields, void *rdata) {
+  struct spg__yield_each_row_stream_data* data = (struct spg__yield_each_row_stream_data *)rdata;
+  VALUE h = rb_hash_new_capa(nfields);
+  VALUE self = data->self;
+  VALUE *colsyms = data->colsyms;
+  VALUE *colconvert= data->colconvert;
+  PGresult *res = pgresult_get(rres);
+  int enc_index = data->enc_index;
+  int j;
+
+  for(j=0; j<nfields; j++) {
+    rb_hash_aset(h, colsyms[j], spg__col_value(self, res, 0, j, colconvert , enc_index));
+  }
+
+  if(data->type == SPG_YIELD_MODEL) {
+    VALUE model = rb_obj_alloc(data->pg_value);
+    rb_ivar_set(model, spg_id_values, h);
+    rb_yield(model);
+  } else {
+    rb_yield(h);
+  }
+  return 1; /* clear the result */
+}
+
 static VALUE spg__yield_each_row_internal(VALUE self, VALUE rconn, VALUE rres, PGresult *res, int enc_index, VALUE *colsyms, VALUE *colconvert) {
-  long nfields;
-  long j;
+  int nfields;
+  int j;
   VALUE h;
   VALUE opts;
   VALUE pg_type;
   VALUE pg_value = Qnil;
   char type = SPG_YIELD_NORMAL;
+  struct spg__yield_each_row_stream_data data;
 
   nfields = PQnfields(res);
 
@@ -1684,8 +1913,20 @@ static VALUE spg__yield_each_row_internal(VALUE self, VALUE rconn, VALUE rres, P
 
   spg_set_column_info(self, res, colsyms, colconvert, enc_index);
 
+  if (spg_use_pg_stream_any) {
+    data.self = self;
+    data.colsyms = colsyms;
+    data.colconvert = colconvert;
+    data.pg_value = pg_value;
+    data.enc_index = enc_index;
+    data.type = type;
+    
+    pgresult_stream_any(rres, spg__yield_each_row_stream, &data);
+    return self;
+  }
+
   while (PQntuples(res) != 0) {
-    h = rb_hash_new();
+    h = rb_hash_new_capa(nfields);
     for(j=0; j<nfields; j++) {
       rb_hash_aset(h, colsyms[j], spg__col_value(self, res, 0, j, colconvert , enc_index));
     }
@@ -1729,7 +1970,7 @@ static VALUE spg__yield_each_row(VALUE self) {
   VALUE rres;
   VALUE rconn;
   int enc_index;
-  long nfields;
+  int nfields;
 
   rconn = rb_ary_entry(self, 1);
   self = rb_ary_entry(self, 0);
@@ -1750,7 +1991,7 @@ static VALUE spg__yield_each_row(VALUE self) {
   else if (nfields <= 1664) return spg__yield_each_row_1664(self, rconn, rres, res, enc_index);
   else {
     rb_funcall(rres, spg_id_clear, 0);
-    rb_raise(rb_eRangeError, "more than 1664 columns in query (%ld columns detected)", nfields);
+    rb_raise(rb_eRangeError, "more than 1664 columns in query (%d columns detected)", nfields);
   }
 
   /* UNREACHABLE */
@@ -1809,10 +2050,21 @@ void Init_sequel_pg(void) {
     }
   }
 
-  if (RTEST(rb_eval_string("defined?(PG::VERSION) && PG::VERSION.to_f >= 1.2"))) {
-    spg_use_pg_get_result_enc_idx = 1;
+  c = rb_eval_string("defined?(PG::VERSION) && PG::VERSION.split('.').map(&:to_i)");
+  if (RB_TYPE_P(c, T_ARRAY) && RARRAY_LEN(c) >= 3) {
+    if (FIX2INT(RARRAY_AREF(c, 0)) > 1) {
+      spg_use_pg_get_result_enc_idx = 1;
+      spg_use_pg_stream_any = 1;
+    } else if (FIX2INT(RARRAY_AREF(c, 0)) == 1) {
+      if (FIX2INT(RARRAY_AREF(c, 1)) >= 2) {
+        spg_use_pg_get_result_enc_idx = 1;
+      }
+      if (FIX2INT(RARRAY_AREF(c, 1)) > 4 || (FIX2INT(RARRAY_AREF(c, 1)) == 4 && FIX2INT(RARRAY_AREF(c, 2)) >= 4)) {
+        spg_use_pg_stream_any = 1;
+      }
+    }
   }
-  
+
   rb_const_set(spg_Postgres, rb_intern("SEQUEL_PG_VERSION_INTEGER"), INT2FIX(SEQUEL_PG_VERSION_INTEGER));
 
   spg_id_BigDecimal = rb_intern("BigDecimal");
@@ -1857,11 +2109,19 @@ void Init_sequel_pg(void) {
   spg_sym_utc = ID2SYM(rb_intern("utc"));
   spg_sym_local = ID2SYM(rb_intern("local"));
   spg_sym_map = ID2SYM(rb_intern("map"));
+  spg_sym_map_array = ID2SYM(rb_intern("map_array"));
+  spg_sym_map_set = ID2SYM(rb_intern("map_set"));
   spg_sym_first = ID2SYM(rb_intern("first"));
   spg_sym_array = ID2SYM(rb_intern("array"));
+  spg_sym_first_array = ID2SYM(rb_intern("first_array"));
+  spg_sym_array_array = ID2SYM(rb_intern("array_array"));
+  spg_sym_first_set = ID2SYM(rb_intern("first_set"));
+  spg_sym_array_set = ID2SYM(rb_intern("array_set"));
   spg_sym_hash = ID2SYM(rb_intern("hash"));
   spg_sym_hash_groups = ID2SYM(rb_intern("hash_groups"));
   spg_sym_model = ID2SYM(rb_intern("model"));
+  spg_sym_all = ID2SYM(rb_intern("all"));
+  spg_sym_all_model = ID2SYM(rb_intern("all_model"));
   spg_sym__sequel_pg_type = ID2SYM(rb_intern("_sequel_pg_type"));
   spg_sym__sequel_pg_value = ID2SYM(rb_intern("_sequel_pg_value"));
 
@@ -1933,7 +2193,7 @@ void Init_sequel_pg(void) {
 
   rb_define_singleton_method(spg_Postgres, "supports_streaming?", spg_supports_streaming_p, 0);
 
-#if HAVE_PQSETSINGLEROWMODE
+#ifdef HAVE_PQSETSINGLEROWMODE
   spg_id_get_result = rb_intern("get_result");
   spg_id_clear = rb_intern("clear");
   spg_id_check = rb_intern("check");
